-- DE.EX: DOS Dialog Editor for Euphoria
-- (c) 1998 David Cuny

without warning
include graphics.e
include get.e
include image.e
include sort.e
include grfont.e

-- load font
    graphic_font()

constant       
    Type    = 1,
    Title   = 2,
    Row     = 3,
    Col     = 4,
    Wide    = 5,
    Deep    = 6,
    Data    = 7,
    Group   = 8,
    Items   = 8
    
constant
    Button  = 1,
    Sle     = 2,
    List    = 3,
    Check   = 4,
    Radio   = 5,
    Label   = 6
    
integer 
    wWide,          -- window width
    wDeep,          -- window depth
    wRow,           -- offset of window row
    wCol            -- offset of window col
sequence
    wTitle          -- window title
    wTitle = "Window Title"
    
constant
    SHIFT_FLAG1_BIT     = 1,    -- status byte for these are at #417
    SHIFT_FLAG2_BIT     = 2,    -- both left and right shifts are tracked
    ALT_FLAG_BIT        = 4     -- at #417


sequence controls


----------------------------------------------------------------------------
function shift_key_state()
        
    -- look for shift key flag (1st or 2nd bit in peek 477H)

    sequence shiftFlag
    
    shiftFlag = int_to_bits( peek( #417 ), 4 )
    if ( shiftFlag[SHIFT_FLAG1_BIT] or shiftFlag[SHIFT_FLAG2_BIT] ) then
        return 1
    else
        return 0
    end if

end function


-----------------------------------------------------------------------------
procedure prompt( sequence text )

    -- display prompt

    -- clear line
    text_color( BLACK )
    bk_color( WHITE )
    position( 1, 1 )
    puts( 1, repeat( ' ', 80 ) )

    -- prompt    
    position( 1, 1 )
    puts( 1, text )
    
end procedure
        

-----------------------------------------------------------------------------
function input( sequence prompt )

    sequence text

    -- clear line
    text_color( BLACK )
    bk_color( WHITE )
    position( 1, 1 )
    puts( 1, repeat( ' ', 80 ) )

    position( 1, 1 )
    puts( 1, prompt & " > " )
    text = gets( 0 )              
    
    return text[1..length(text)-1]
    
end function        

-----------------------------------------------------------------------------    
procedure w_position( integer row, integer col)

    -- position relative to the window
    position( wRow+row, wCol+col)
    
end procedure

-----------------------------------------------------------------------------
function center( sequence text, integer wide )
    
    -- return offset to center text
    return floor((wide/2) - (length( text )/2))

end function




-----------------------------------------------------------------------------
procedure draw_control( integer active, integer this )

    integer tmp
    sequence text

    -- in range?
    if this > length( controls ) then
        return
    end if

    -- set colors
    if active then        
        bk_color( BLACK )
        text_color( WHITE )
    else
        bk_color( WHITE )
        text_color( BLACK )
    end if
        
    text = controls[this][Title]

    if controls[this][Type] = Button then
        -- must be at least as wide as text
        if controls[this][Wide] < length( controls[this][Title] ) then
            controls[this][Wide] = length( controls[this][Title] )
        end if
        
        -- center
        tmp = floor((controls[this][Wide]-length( controls[this][Title] )) / 2)
        -- convert to spaces
        text = repeat( ' ', tmp )          
        -- build button
        if active then
            text = 16 & text & controls[this][Title] & text & 17
        else
            text = ' ' & text & controls[this][Title] & text & ' '
        end if
        

        -- this requires special colors, so display here and exit
        w_position( controls[this][Row], controls[this][Col] )    
        put_colors( {BLACK, BRIGHT_WHITE}, text )
        return

    
    elsif controls[this][Type] = Check then
        -- show setting
        if compare( controls[this][Data], "1" ) = 0 then
            text = CHECK_ON
        else
            text = CHECK_OFF
        end if
        
        text = text & ' ' & controls[this][Title]

    elsif controls[this][Type] = Radio then
        -- show data
        if compare( controls[this][Data], "1" ) = 0 then
            text = RADIO_ON
        else
            text = RADIO_OFF
        end if
        
        text = text & ' ' & controls[this][Title]
        
          
    elsif controls[this][Type] = Label then
        w_position( controls[this][Row], controls[this][Col] )
        puts( 1, controls[this][Title] )
    
    elsif controls[this][Type] = List then
        -- draw box
        
        -- top
        tmp = controls[this][Row]
        w_position( tmp+1, controls[this][Col] )
        puts( 1, 218 & repeat( 196, controls[this][Wide] ) & 191 )                    
        
        -- middle
        for i = 2 to controls[this][Deep]+1 do
            w_position( tmp+i, controls[this][Col] )
            puts( 1, 179 & repeat( ' ', controls[this][Wide] ) & 179 )
        end for

        -- bottom
        w_position( tmp+controls[this][Deep]+2, controls[this][Col] )
        puts( 1, 192 & repeat( 196, controls[this][Wide] ) & 217 )

        -- title
        text = " " & controls[this][Title]
        
    elsif controls[this][Type] = Sle then
        -- get data
        text = controls[this][Data]
        
        -- strip quote
        if text[1] = '\"' then
            text = text[2..length(text)-1]
        end if
        
        -- clip
        if length( text ) > controls[this][Wide] then
            -- trim
            text = text[1..controls[this][Wide]]
        else
            -- pad
            text = text & repeat( ' ', controls[this][Wide] - length( text ) )
        end if

        -- position of frame
        tmp = controls[this][Col] + length( controls[this][Title] ) + 1

        -- middle
        w_position( controls[this][Row], tmp )
        puts( 1, "[" & text & "]" )

        -- title
        text = controls[this][Title]
          
    elsif controls[this][Type] = Sle then
        
    end if
    
    w_position( controls[this][Row], controls[this][Col] )    
    puts( 1, text )
    
end procedure


----------------------------------------------------------------------------- 
procedure position_buttons( integer wide, integer deep )

    integer totalWide, count, space, tmp

    -- position the buttons in the button portion of the window
    
    -- calculate the space taken                            
    count = 0
    totalWide = 0
    for i = 1 to length( controls ) do
        if controls[i][Type] = Button then
            count = count + 1
            totalWide = totalWide + controls[i][Wide] + 4
        end if        
    end for
    
    -- any buttons?

    if count > 0 then
        -- divide evenly        
        space = floor( (( wide - totalWide ) / (count+1)) )
        tmp = 2 + space
        for i = 1 to length( controls ) do
            if controls[i][Type] = Button then
                controls[i][Col] = tmp
                controls[i][Row] = deep-1
                tmp = tmp + controls[i][Wide] + 4 + space
            end if        
        end for
    end if

end procedure


-----------------------------------------------------------------------------    
procedure draw_window()

    integer tmp             

    bk_color( BLUE )
    clear_screen()

    -- set values
    wRow = floor( (25 - wDeep)/2 ) - 1
    wCol = floor( (80 - wWide)/2 ) - 1

    -- set colors
    bk_color( WHITE )
    text_color( BLACK )

    -- draw top
    w_position( 0, 0 )
    put_colors( {BLACK, BRIGHT_WHITE}, repeat( ' ', wWide ) )
    w_position( 0, center( wTitle & " ", wWide ) )
    put_colors( {BLACK, BRIGHT_WHITE}, ' ' & wTitle & ' ' )
    -- body of window    
    for i = 1 to wDeep do
        w_position( i, 0 )
        puts( 1, repeat( ' ', wWide ) )
    end for    

    -- resize the buttons
    position_buttons( wWide, wDeep )

end procedure

-----------------------------------------------------------------------------
procedure draw_all()
    
    -- draw window and all controls

    draw_window()
    for i = 1 to length(controls) do
        draw_control( 0, i )
    end for
    
end procedure


-----------------------------------------------------------------------------
function menu( sequence s )
    
    integer pick, key

    pick = 1
    
    while 1 do
    
        text_color( BLACK )
        bk_color( WHITE )
        position( 1, 1 )
        puts( 1, repeat( ' ', 80 ) )
        
        position( 1, 1 )    
        for i = 1 to length( s ) do
            if pick = i then
                -- reverse colors
                text_color( WHITE )
                bk_color( BLACK )
            else
                -- normal colors
                text_color( BLACK )
                bk_color( WHITE )
            end if
                
            printf( 1, " %s ", {s[i]} )                
        end for

        key = wait_key()            
        if key = 27 then
            -- escape
            return 0
        
        elsif key = 13 then
            -- enter
            return pick
        
        elsif key = 331 then
            -- left
            pick = pick - 1
            if pick < 1 then
                pick = length(s)
            end if
        
        elsif key = 333 then
            -- right
            pick = pick + 1
            if pick > length(s) then
                pick = 1
            end if
        
        else
            position( 2, 1 )
            printf( 1, "%d", {key} )
        end if
        
    end while

end function


-----------------------------------------------------------------------------
procedure add_control()

    integer pick
    sequence control

    pick = menu( {"Button", "SLE", "List", "Check", "Radio", "Label"})
    
    -- exit?
    if pick = 0 then 
        return
    end if  

    -- add control                             
    control = repeat( 0, Items )
    control[Type] = "Type"
    control[Row] = 1
    control[Col] = 1
    control[Wide] = 1
    control[Deep] = 0
    control[Data] = "0"
    
    control[Type] = pick
    
    if pick = Button then
        control[Title] = "Button"        
        
    elsif pick = Check then        
        control[Title] = "Check Box"
    
    elsif pick = Label then        
        control[Title] = "Label"
        
    elsif pick = List then        
        control[Title] = "List"
        control[Data] = "{}"
        control[Wide] = 8
        control[Deep] = 1
        
    elsif pick = Radio then        
        control[Title] = "Radio"
    
    elsif pick = Sle then        
        control[Title] = "Sle"
        control[Wide] = 8
        control[Data] = "\"\""
        
    end if

    controls = append( controls, control )
    
end procedure                



-----------------------------------------------------------------------------
procedure sort_controls()

    -- sort the controls           

    -- add sort values
    for i = 1 to length( controls ) do
        -- weight by position
        controls[i] = { (controls[i][Row]*100)+controls[i][Col], 
                            controls[i] }
    end for
    
    controls = sort( controls )
    
    -- remove sort values
    for i = 1 to length( controls ) do
        controls[i] = controls[i][2]
    end for

end procedure

-----------------------------------------------------------------------------
procedure edit_controls()

    integer key, this, shiftFlag, unknownKey
    sequence oldControls, screen, text

    this = 1
    unknownKey = 0

    while 1 do    
        
        -- draw the window
        draw_window()

        -- draw all but active controls
        for i = 1 to length( controls ) do
            if i != this then   
                draw_control( 0, i )
            end if        
        end for

        -- draw active control
        if length( controls ) then
            draw_control( 1, this )
        end if

        -- prompt
        prompt( "Tab=Navigate  Arrows=Move  Enter=Rename  Insert=Add  Delete=Delete" )

        -- prior key
        if unknownKey then
            -- debugging
            -- position( 2, 1 )
            -- printf( 1, "%d", {unknownKey} )
            unknownKey = 0        
        end if
        
        key = wait_key()
        shiftFlag = shift_key_state()

        if key = 9 then
            -- tab
            this = this + 1
            if this > length( controls ) then
                this = 1
            end if

        elsif key = 10 then
            -- ctrl+Enter
            text = input( "Window title" )
            if length( text ) then
                wTitle = text
            end if
            
            draw_all()
        
        elsif key = 271 then
            -- back tab
            this = this - 1
            if this < 1 then
                this = length( controls )
            end if
            
        elsif key = 13 
        and   length(controls ) then
            -- enter
            text =input( "New Control Title" )
            if length( text ) then
                controls[this][Title] = text
            end if
                
        
        elsif key = 27 then
            -- escape
            return

        elsif key = ' '
        and   length( controls ) then

            -- enter value
            text = input( "Value for " & controls[this][Title] )
            if length( text ) then
                -- set to value
                controls[this][Data] = text

                -- longer than control?
                if text[1] = '\"' then                                    
                    if length( text )-2 > controls[this][Wide] then
                        -- increase size
                        controls[this][Wide] = length( text )-2
                    end if
                end if
            end if
            
            -- button group?
            if controls[this][Type] = Radio then
                text = input( "Group for " & controls[this][Title] )
                if length( text ) then
                
                    -- attempt to convert
                    text = value( text )                                        
                    if text[1] = GET_SUCCESS then
                        -- set to value
                        controls[this][Group] = text[2]
                    end if
                end if
            end if
            

        elsif key = 327
        and length( controls ) then
            sort_controls()
            this = 1
            draw_all()
        
        elsif key = 328 
        and   length( controls ) then            
            -- up
            if shiftFlag then    
                -- shrink height                                               
                if controls[this][Deep] > 1 then
                    controls[this][Deep] = controls[this][Deep] - 1
                end if
            else
                controls[this][Row] = controls[this][Row] - 1
            end if

        elsif key = 336 
        and   length( controls ) then
            -- down                  
            if shiftFlag then
                -- increase height                                               
                controls[this][Deep] = controls[this][Deep] + 1
                
            else
                -- move control
                controls[this][Row] = controls[this][Row] + 1
            
            end if

    
        elsif key = 331 
        and   length( controls ) then
            -- left
            if shiftFlag then
                -- shrink width
                if controls[this][Wide] > 1 then
                    controls[this][Wide] = controls[this][Wide] - 1
                end if
            
            else
                -- move control
                controls[this][Col] = controls[this][Col] - 1
            
            end if
        
        elsif key = 333 then
            -- right    
            if shiftFlag then
                -- increase width
                controls[this][Wide] = controls[this][Wide] + 1
            
            else
                -- move control
                controls[this][Col] = controls[this][Col] + 1

            end if
            
        elsif key = 338 then
            -- insert
            add_control()
            this = length( controls )

        elsif key = 339 then
            -- delete
            controls = controls[1..this-1] & 
                    controls[this+1..length(controls)]
            this = 1

        elsif key = 371 then
            -- ctrl left
            if wWide > 1 then
                wWide = wWide - 1
            end if
            
        elsif key = 372 then
            -- ctrl right
            wWide = wWide + 1
                                            
        elsif key = 397 then
            -- ctrl up
            if wDeep > 1 then
                wDeep = wDeep - 1
            end if

        elsif key = 401 then
            -- ctrl down 
            wDeep = wDeep + 1

        else
            unknownKey = key
        end if
        
    end while    
    
end procedure


-----------------------------------------------------------------------------
integer handle
sequence typeName, control

-- Defaults
    controls = {}
    wWide = 30
    wDeep = 10

    -- define ok button
    control = repeat( 0, 7 )
    control[Type] = Button
    control[Title] = "OK"
    control[Row] = 0
    control[Col] = 0
    control[Wide] = 6
    control[Deep] = 0
    control[Data] = "0"    
    controls = append( controls, control )

    -- define cancel button    
    control[Title] = "Cancel"
    control[Wide] = 8
    controls = append( controls, control )
    


-- edit
    edit_controls()   
    sort_controls()
    
    
    
-- generate code
    handle = open( "dcode.ex", "w" )
    puts( handle, "-- Dialog code\n" )
    puts( handle, "-- Generated by Dialog Editor v1.6\n" )
    puts( handle, "-- (c) 1997 David Cuny\n" )
    puts( handle, "\n\n" )         
    puts( handle, "-- include files\n" )    
    puts( handle, "\tinclude dialog.e\n" )
    puts( handle, "\n\n" )     
    puts( handle, "-- for returned result\n" )
    puts( handle, "\tobject result\n" )
    puts( handle, "\n\n" )
    puts( handle, "function user_dialog()\n" )    
    puts( handle, "\n" )
    puts( handle, "\t-- new dialog\n" )
    puts( handle, "\t\tnew_dialog()\n" )
    puts( handle, "\n" )             

    -- controls
    puts( handle, "\t-- create controls\n" )
    for i = 1 to length( controls ) do
        if controls[i][Type] != Button then
        
            if      controls[i][Type] = Check   then typeName = "Check"
            elsif   controls[i][Type] = Label   then typeName = "Label"
            elsif   controls[i][Type] = List    then typeName = "List"
            elsif   controls[i][Type] = Radio   then typeName = "Radio"
            elsif   controls[i][Type] = Sle     then typeName = "Sle"
            end if  
            
            if controls[i][Type] = Radio then
            
                -- create a radio control
                printf( handle, 
                    "\t\tadd_radio( \"%s\", %d, %d, %d, %s )\n",
                    {controls[i][Title],                    
                     controls[i][Row],
                     controls[i][Col],
                     controls[i][Group],
                     controls[i][Data]} )

            else                     
            
                printf( handle,                      
                    "\t\tadd_control( %s, \"%s\", %d, %d, %d, %d, %s )\n",
                    {typeName, 
                     controls[i][Title],
                     controls[i][Row],
                     controls[i][Col],
                     controls[i][Wide],
                     controls[i][Deep],
                     controls[i][Data]} )
                  
            end if
        end if
    end for
    
    -- buttons
    puts( handle, "\n" )
    puts( handle, "\t-- create buttons\n" )    
    for i = 1 to length( controls ) do
        if controls[i][Type] = Button then
            printf( handle, "\t\tadd_button( \"%s\", %d )\n",
                    {controls[i][Title], controls[i][Wide]} )
        end if            
    end for

    -- window         
    puts( handle, "\n" )          
    printf( handle, "\t\treturn run_dialog( %d, %d, \"%s\" )\n", 
                {wWide, wDeep, wTitle} )
    
    puts( handle, "\n" )    
    puts( handle, "end function\n" )    
    puts( handle, "\n\n" )    
    puts( handle, "-- run dialog\n" )    
    puts( handle, "\tresult = user_dialog()\n" )    
    
                
-- close file
    close( handle )

-- restore fonts
    reset_font()

-- clear screen
    text_color( WHITE )
    bk_color( BLACK )
    clear_screen()
    puts( 1, "File DCODE.EX has been created.\n" )

